"""Functions for parsing `Duration`s from strings."""

from stuom.duration import (
    Duration,
    Hours,
    HundredNanoseconds,
    Microseconds,
    Milliseconds,
    Minutes,
    Nanoseconds,
    Seconds,
)


def find_minimal_si(d: Duration) -> Duration:
    """Find the given duration's minimum-order SI type that minimizes the number of digits needed
    to represent the given value.

    Args:
        d (Duration): The duration to parse

    Returns:
        Duration: A duration that minimizes the number of digits needed to represent the given
        value.
    """
    duration_us = d.convert_duration(Microseconds)
    duration_ms = d.convert_duration(Milliseconds)
    if duration_ms >= Milliseconds(1000):
        return d.convert_duration(Seconds)

    if duration_us >= Microseconds(1000):
        return d.convert_duration(Milliseconds)

    else:
        return duration_us


def parse_minimal_si(d: Duration) -> str:
    """String format a duration into its minimum-order SI duration.

    Args:
        d (Duration): The duration to process.

    Returns:
        str: The minimal subclass as a string.
    """
    return str(find_minimal_si(d))


def parse_duration(ts_str: str) -> Duration:
    """Parse a `Duration` from a string."""
    if len(ts_str) <= 1:
        raise ValueError(f"Invalid timestamp string: {ts_str}")

    ts_str = ts_str.lower()
    if "hr" in ts_str:
        return Hours(ts_str.replace("hrs", "").replace("hr", ""))

    if "min" in ts_str:
        return Minutes(ts_str.replace("mins", "").replace("min", ""))

    elif "ms" in ts_str:
        return Milliseconds(ts_str.replace("ms", ""))

    elif "us" in ts_str or "μs" in ts_str:
        return Microseconds(ts_str.replace("us", "").replace("μs", ""))

    elif "hns" in ts_str:
        return HundredNanoseconds(ts_str.replace("hns", ""))

    elif "ns" in ts_str:
        return Nanoseconds(ts_str.replace("ns", ""))

    # Note: Order is important since we want to run through all of the above cases before reaching
    # "s" otherwise this would match first.
    elif "s" in ts_str:
        return Seconds(ts_str.replace("s", ""))

    raise ValueError("No timestamp type specified!")



"""Simple statically-typed unit of measurements with order conversions."""

from abc import abstractmethod
from typing import SupportsFloat, TypeAlias, TypeVar

HAS_PYDANTIC_CORE = False
"""Whether the environment is configured with the pydantic_core package installed."""

# Attempt to import pydantic_core, and determine whether validation support is to be added or not.
try:
    from collections.abc import Callable
    from typing import Any

    from pydantic_core import core_schema

    HAS_PYDANTIC_CORE = True
except ImportError:
    pass


SiT = TypeVar("SiT", bound="HasSiOrder")
IntT: TypeAlias = int | float


# SI unit types
class HasSiOrder(float):
    """Represents a unit of measurement type that has an SI order at the type level."""

    @staticmethod
    @abstractmethod
    def order() -> float:
        raise NotImplementedError()

    def convert_si(self, to_cls: type[SiT]) -> SiT:
        return _convert_si(self, to_cls)

    @classmethod
    def from_si(cls: type[SiT], other: "HasSiOrder") -> SiT:
        return other.convert_si(cls)

    def __mul__(self: SiT, value: IntT) -> SiT:
        self_type = type(self)
        return self_type(float.__mul__(self, value))

    def __add__(self: SiT, value: IntT) -> SiT:
        self_type = type(self)
        return self_type(float.__add__(self, value))

    def __sub__(self: SiT, value: IntT) -> SiT:
        self_type = type(self)
        return self_type(float.__sub__(self, value))

    def __truediv__(self: SiT, value: IntT) -> SiT:
        self_type = type(self)
        return self_type(float.__truediv__(self, value))

    def __floordiv__(self: SiT, value: IntT) -> SiT:
        self_type = type(self)
        return self_type(float.__floordiv__(self, value))

    if HAS_PYDANTIC_CORE:

        @classmethod
        def validate(cls, value: object, _):
            if not isinstance(value, SupportsFloat):
                raise TypeError(f"Unsupported standard unit number type: {type(value)}")

            return cls(float(value))

        @classmethod
        def __get_pydantic_json_schema__(
            cls,
            schema: core_schema.CoreSchema,  # type: ignore
            handler: Callable[[Any], core_schema.CoreSchema],  # type: ignore
        ) -> core_schema.CoreSchema:  # type: ignore
            return handler(core_schema.float_schema())

        @classmethod
        def __get_pydantic_core_schema__(cls, source, handler):
            return core_schema.with_info_plain_validator_function(cls.validate)


def _convert_si(from_value: HasSiOrder, to_cls: type[SiT]) -> SiT:
    return to_cls(from_value * 10 ** (to_cls.order() - from_value.order()))



"""Units of measurement base class for working with computer memory sizes."""

from typing import TypeVar

from stuom.uom import HasSiOrder

MemorySizeT = TypeVar("MemorySizeT", bound="MemorySize")


class MemorySize(HasSiOrder):
    """Represents a byte-based unit of measuring memory size.

    Note that there are two standards for describing computer memory size. One is the International
    System of Units (SI) definition, in which the decimal number system is used: 1000 bytes is
    kilobytes (kB). The other is binary-based, in where 1024 bytes is 1 KB. The IEC recommends the
    use of Kibi as a prefix in place of Kilo however. We intend to follow the IEC80000-13
    international standard on this matter and distinguish KiloBytes vs. KibiBytes.
    """

    def convert_memmory_size(self, to_cls: type[MemorySizeT]) -> MemorySizeT:
        return self.convert_si(to_cls)

    @classmethod
    def from_memory_size(cls: type[MemorySizeT], other: "MemorySize") -> MemorySizeT:
        return other.convert_memmory_size(cls)


class Bytes(MemorySize):
    """8 bits."""

    def __str__(self) -> str:
        return super().__str__() + " B"

    @staticmethod
    def order() -> int:
        return 0



"""Contains computer memory size units of memory size following the EIC decimal convention."""

from stuom.memory_size.memory_size import MemorySize


class KiloBytes(MemorySize):
    """1_000 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " kB"

    @staticmethod
    def order() -> int:
        return -3


class MegaBytes(MemorySize):
    """1_000_000 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " MB"

    @staticmethod
    def order() -> int:
        return -6


class GigaBytes(MemorySize):
    """1_000_000_000 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " GB"

    @staticmethod
    def order() -> int:
        return -9


class TeraBytes(MemorySize):
    """1_000_000_000_000 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " TB"

    @staticmethod
    def order() -> int:
        return -12


class PetaBytes(MemorySize):
    """1_000_000_000_000 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " PB"

    @staticmethod
    def order() -> int:
        return -15



"""Units of measurement for working with durations."""

import math
from typing import TypeVar

from stuom.uom import HasSiOrder

DurationT = TypeVar("DurationT", bound="Duration")


class Duration(HasSiOrder):
    """Represents a duration and is functionally equivalent to its base."""

    def convert_duration(self, to_cls: type[DurationT]) -> DurationT:
        return self.convert_si(to_cls)

    @classmethod
    def from_duration(cls, other: "Duration"):
        return other.convert_duration(cls)


class Hours(Duration):
    """hrs."""

    def __str__(self) -> str:
        return super().__str__() + " hrs"

    @staticmethod
    def order() -> float:
        return -math.log(3600, 10)


class Minutes(Duration):
    """mins."""

    def __str__(self) -> str:
        return super().__str__() + " mins"

    @staticmethod
    def order() -> float:
        return -(math.log(6, 10) + 1)


# Seconds-based SI units
class Seconds(Duration):
    """s."""

    def __str__(self) -> str:
        return super().__str__() + "s"

    @staticmethod
    def order() -> float:
        return 0


class Milliseconds(Duration):
    """ms."""

    def __str__(self) -> str:
        return super().__str__() + "ms"

    @staticmethod
    def order() -> float:
        return 3


class Microseconds(Duration):
    """us."""

    def __str__(self) -> str:
        return super().__str__() + "us"

    @staticmethod
    def order() -> float:
        return 6


class HundredNanoseconds(Duration):
    """hns."""

    def __str__(self) -> str:
        return super().__str__() + "hns"

    @staticmethod
    def order() -> float:
        return 7


class Nanoseconds(Duration):
    """ns."""

    def __str__(self) -> str:
        return super().__str__() + "ns"

    @staticmethod
    def order() -> float:
        return 9



"""Contains units of measurement for working with electricity units such as power."""

from typing import TypeVar

from stuom.uom import HasSiOrder

PotentialT = TypeVar("PotentialT", bound="ElectricPotential")


class ElectricPotential(HasSiOrder):
    """A unit of measurement representing electric potential."""

    def convert_potential(self, to_cls: type[PotentialT]) -> PotentialT:
        return self.convert_si(to_cls)

    @classmethod
    def from_potential(cls: type[PotentialT], other: "ElectricPotential") -> PotentialT:
        return other.convert_potential(cls)


class Kilovolts(ElectricPotential):
    """kV."""

    def __str__(self) -> str:
        return super().__str__() + "kV"

    @staticmethod
    def order() -> int:
        return -3


class Volts(ElectricPotential):
    """V."""

    def __str__(self) -> str:
        return super().__str__() + "V"

    @staticmethod
    def order() -> int:
        return 0


class Millivolts(ElectricPotential):
    """mV."""

    def __str__(self) -> str:
        return super().__str__() + "mV"

    @staticmethod
    def order() -> int:
        return 3


PowerT = TypeVar("PowerT", bound="Power")


class Power(HasSiOrder):
    """The amount of energy transferred per unit time."""

    def convert_power(self, to_cls: type[PowerT]) -> PowerT:
        return self.convert_si(to_cls)

    @classmethod
    def from_power(cls: type[PowerT], other: "Power") -> PowerT:
        return other.convert_power(cls)


class Watts(Power):
    """W."""

    def __str__(self) -> str:
        return super().__str__() + "W"

    @staticmethod
    def order() -> int:
        return 0

    @staticmethod
    def from_current_and_potential(
        current: "Current", potential: ElectricPotential
    ) -> "Watts":
        return Watts(Amps.from_si(current) * Volts.from_si(potential))


class Deciwatts(Power):
    """dW."""

    def __str__(self) -> str:
        return super().__str__() + "dW"

    @staticmethod
    def order() -> int:
        return 1


class Milliwatts(Power):
    """mW."""

    def __str__(self) -> str:
        return super().__str__() + "mW"

    @staticmethod
    def order() -> int:
        return 3


CurrentT = TypeVar("CurrentT", bound="Current")


class Current(HasSiOrder):
    """A unit of measurement representing electric current."""

    def convert_current(self, to_cls: type[CurrentT]) -> CurrentT:
        return self.convert_si(to_cls)

    @classmethod
    def from_current(cls: type[CurrentT], other: "Current") -> CurrentT:
        return other.convert_current(cls)


class Amps(Current):
    """A."""

    def __str__(self) -> str:
        return super().__str__() + "A"

    @staticmethod
    def order() -> int:
        return 0


class Milliamps(Current):
    """mA."""

    def __str__(self) -> str:
        return super().__str__() + "mA"

    @staticmethod
    def order() -> int:
        return 3


class Microamps(Current):
    """uA."""

    def __str__(self) -> str:
        return super().__str__() + "uA"

    @staticmethod
    def order() -> int:
        return 6



"""Test unit of measurement types."""

import math

from stuom import (
    Angstroms,
    Deciwatts,
    Hours,
    HundredNanoseconds,
    Kilovolts,
    Meters,
    Microamps,
    Microseconds,
    Milimeters,
    Milliseconds,
    Minutes,
    Nanoseconds,
    ReciprocalCentimeters,
    ReciprocalMeters,
    Seconds,
    Watts,
)
from stuom.electricity import Amps, Volts
from stuom.length import Centimeters, Micrometers


def test_uom_operators():
    mins_sum = Minutes(2) + 4
    assert isinstance(mins_sum, Minutes)
    assert mins_sum == Minutes(6)

    mins_prod = Minutes(2) * 4
    assert isinstance(mins_prod, Minutes)
    assert mins_prod == Minutes(8)

    mins_truediv = Minutes(4) / 2
    assert isinstance(mins_truediv, Minutes)
    assert mins_truediv == Minutes(2)

    mins_floordiv = Minutes(4.5) // 2
    assert isinstance(mins_floordiv, Minutes)
    assert mins_floordiv == Minutes(2)

    mins_sub = Minutes(9) - 2
    assert isinstance(mins_sub, Minutes)
    assert mins_sub == Minutes(7)


def test_duration_uom():
    assert math.isclose(Minutes.from_duration(Hours(2)), Minutes(60 * 2))
    assert math.isclose(Hours(2).convert_duration(Minutes), Minutes(60 * 2))
    assert math.isclose(Minutes.from_duration(Seconds(10)), Minutes(1 / 6))
    assert math.isclose(Seconds.from_duration(Minutes(10)), 10 * 60)
    assert math.isclose(Milliseconds.from_duration(Minutes(10)), 10 * 60 * 1000)
    assert math.isclose(Microseconds.from_duration(Minutes(10)), 10 * 60 * 1000 * 1000)
    assert math.isclose(
        Nanoseconds.from_duration(Minutes(10)), 10 * 60 * 1000 * 1000 * 1000
    )
    assert math.isclose(
        HundredNanoseconds.from_duration(Minutes(10)), 10 * 60 * 1000 * 1000 * 10
    )


def test_length_uom():
    assert math.isclose(Meters(0.005), Milimeters(5).convert_length(Meters))
    assert math.isclose(Meters(1), Meters.from_length(Milimeters(1000)))
    assert math.isclose(Meters(2), Meters.from_length(Angstroms.from_length(Meters(2))))
    assert math.isclose(Meters(2).convert_length(Angstroms), Angstroms(2e10))
    assert math.isclose(Angstroms(2e10).convert_length(Meters), Meters(2))

    assert math.isclose(
        ReciprocalMeters(10e-2)
        .convert_from_reciprocal_length(ReciprocalCentimeters)
        .convert_from_reciprocal_length(ReciprocalMeters),
        ReciprocalMeters(10e-2),
    )

    assert math.isclose(
        ReciprocalMeters(1).convert_from_reciprocal_length(ReciprocalCentimeters),
        ReciprocalCentimeters(100),
    )

    assert math.isclose(ReciprocalMeters(10e-2).to_meters(), Meters(1 / 10e-2))


def test_electricity_uom():
    power = Watts.from_current_and_potential(Microamps(225), Kilovolts(20))
    assert power == Watts(4.5)
    assert Deciwatts.from_power(power) == 45
    assert Volts.from_potential(Kilovolts(20)) == Volts(20000)
    assert Amps.from_current(Microamps(225)) == Amps(0.000225)


def test_uom_str_representations():
    # Length
    assert str(Micrometers(10)) == "10.0um"
    assert str(Milimeters(10)) == "10.0mm"
    assert str(Meters(10)) == "10.0m"
    assert str(Centimeters(10)) == "10.0cm"

    # Electricity
    assert str(Microamps(10)) == "10.0uA"
    assert str(Amps(10)) == "10.0A"

    assert str(Watts(10)) == "10.0W"
    assert str(Deciwatts(10)) == "10.0dW"

    assert str(Volts(10)) == "10.0V"
    assert str(Kilovolts(10)) == "10.0kV"

    # Duration
    assert str(Nanoseconds(10)) == "10.0ns"
    assert str(Microseconds(10)) == "10.0us"
    assert str(Seconds(10)) == "10.0s"
    assert str(Minutes(10)) == "10.0 mins"
    assert str(Hours(10)) == "10.0 hrs"



"""Contains computer memory size units of memory size following the EIC binary convention.

The current implementation of these units incurs some added floating-point error. The reason this
occurs is simply because the uom.py module implements SI conversions using a base-10 exponent.
"""

import math

from stuom.memory_size.memory_size import MemorySize


class KibiBytes(MemorySize):
    """2^10 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " kiB"

    @staticmethod
    def order() -> float:
        return -math.log(2**10, 10)


class MebiBytes(MemorySize):
    """2^20 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " MiB"

    @staticmethod
    def order() -> float:
        return -math.log(2**20, 10)


class GibiBytes(MemorySize):
    """2^30 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " GiB"

    @staticmethod
    def order() -> float:
        return -math.log(2**30, 10)


class TebiBytes(MemorySize):
    """2^40 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " TiB"

    @staticmethod
    def order() -> float:
        return -math.log(2**40, 10)


class PebiBytes(MemorySize):
    """2^50 bytes."""

    def __str__(self) -> str:
        return super().__str__() + " PiB"

    @staticmethod
    def order() -> float:
        return -math.log(2**50, 10)



"""Test that we can create basic pydantic models using units of measurement."""

import pytest
from stuom.duration import Seconds
from stuom.electricity import Kilovolts

try:
    from pydantic import BaseModel  # type: ignore

except (ImportError, ModuleNotFoundError):
    pytest.skip(allow_module_level=True)


def test_duration_fields_work_in_pydantic_models():
    """This ensures that one can construct pydantic `BaseModel`s with the `Duration` units of
    measurement.
    """

    class TestModel(BaseModel):
        duration: Seconds

    _ = TestModel(duration=Seconds(2))


def test_uom_floatlike_validates():
    class PydanticModel(BaseModel):
        voltage: Kilovolts

    model = PydanticModel.model_validate({"voltage": 10})
    assert model.voltage == 10

def test_uom_non_floatlike_fails_to_validate():
    class PydanticModel(BaseModel):
        voltage: Kilovolts

    with pytest.raises(TypeError):
        _ = PydanticModel.model_validate({"voltage": PydanticModel})



"""Statically-typed units of measurement."""

from . import duration, electricity, length, parse, uom
from .duration import (
    Duration,
    Hours,
    HundredNanoseconds,
    Microseconds,
    Milliseconds,
    Minutes,
    Nanoseconds,
    Seconds,
)
from .electricity import (
    Current,
    Deciwatts,
    ElectricPotential,
    Kilovolts,
    Microamps,
    Milliamps,
    Millivolts,
    Power,
    Volts,
    Watts,
)
from .length import (
    Angstroms,
    Centimeters,
    Meters,
    Micrometers,
    Milimeters,
    Nanometers,
    ReciprocalCentimeters,
    ReciprocalLength,
    ReciprocalMeters,
    ReciprocalMicrometers,
    ReciprocalMilimeters,
)
from .memory_size import (
    Bytes,
    GibiBytes,
    GigaBytes,
    KibiBytes,
    KiloBytes,
    MebiBytes,
    MegaBytes,
    MemorySize,
    PebiBytes,
    PetaBytes,
    TebiBytes,
    TeraBytes,
)



"""Test unit of measurement parsing."""

import pytest
from stuom.duration import (
    Hours,
    HundredNanoseconds,
    Microseconds,
    Milliseconds,
    Minutes,
    Nanoseconds,
    Seconds,
)
from stuom.parse import find_minimal_si, parse_duration, parse_minimal_si


def test_parse_minimal_si():
    assert type(find_minimal_si(Milliseconds(0.05))) is Microseconds
    assert type(find_minimal_si(Milliseconds(5000))) is Seconds
    assert type(find_minimal_si(Milliseconds(5))) is Milliseconds
    assert parse_minimal_si(Milliseconds(0.05)) == "50.0us"
    assert parse_minimal_si(Milliseconds(5000)) == "5.0s"
    assert parse_minimal_si(Seconds(0.01)) == "10.0ms"


def test_parse():
    assert parse_duration("10hrs") == Hours(10)
    assert parse_duration("10mins") == Minutes(10)
    assert parse_duration("10s") == Seconds(10)
    assert parse_duration("10ms") == Milliseconds(10)
    assert parse_duration("10us") == Microseconds(10)
    assert parse_duration("10hns") == HundredNanoseconds(10)
    assert parse_duration("10ns") == Nanoseconds(10)

    with pytest.raises(ValueError):
        parse_duration("10x")

    with pytest.raises(ValueError):
        parse_duration("1")



"""Units of measurement for working with computer memory sizes."""

from .binary_memory_size import (
    GibiBytes,
    KibiBytes,
    MebiBytes,
    PebiBytes,
    TebiBytes,
)
from .decimal_memory_size import (
    GigaBytes,
    KiloBytes,
    MegaBytes,
    PetaBytes,
    TeraBytes,
)
from .memory_size import Bytes, MemorySize



"""Units of measurement for working with lengths."""

from typing import TypeVar

from stuom.uom import HasSiOrder

LengthT = TypeVar("LengthT", bound="Length")


class Length(HasSiOrder):
    """A general-purpose SI-based length."""

    def convert_length(self, to_cls: type[LengthT]) -> LengthT:
        return self.convert_si(to_cls)

    @classmethod
    def from_length(cls: type[LengthT], other: "Length") -> LengthT:
        return other.convert_length(cls)


class Meters(Length):  # noqa: D101
    """m."""

    def __str__(self) -> str:  # noqa: D105
        return super().__str__() + "m"

    @staticmethod
    def order() -> int:
        return 0


class Centimeters(Length):
    """cm."""

    def __str__(self) -> str:
        return super().__str__() + "cm"

    @staticmethod
    def order() -> int:
        return 2


class Milimeters(Length):
    """mm."""

    def __str__(self) -> str:
        return super().__str__() + "mm"

    @staticmethod
    def order() -> int:
        return 3


class Micrometers(Length):
    """um."""

    def __str__(self) -> str:
        return super().__str__() + "um"

    @staticmethod
    def order() -> int:
        return 6


class Nanometers(Length):
    """nm."""

    def __str__(self) -> str:
        return super().__str__() + "nm"

    @staticmethod
    def order() -> int:
        return 9


class Angstroms(Length):
    """lambda."""

    @staticmethod
    def order() -> int:
        return 10


ReciprocalLengthT = TypeVar("ReciprocalLengthT", bound="ReciprocalLength")


class ReciprocalLength(HasSiOrder):
    """1 / `Length`."""

    def convert_from_reciprocal_length(
        self, to_cls: type[ReciprocalLengthT]
    ) -> ReciprocalLengthT:
        return self.convert_si(to_cls)

    @classmethod
    def from_reciprocal_length(
        cls: type[ReciprocalLengthT], other: "ReciprocalLength"
    ) -> ReciprocalLengthT:
        return other.convert_from_reciprocal_length(cls)


class ReciprocalMeters(ReciprocalLength):
    """m^-1."""

    @staticmethod
    def order() -> int:
        return 0

    def to_meters(self) -> Meters:
        return Meters(1 / self)


class ReciprocalCentimeters(ReciprocalLength):
    """cm^-1."""

    @staticmethod
    def order() -> int:
        return 2


class ReciprocalMilimeters(ReciprocalLength):
    """mm^-1."""

    @staticmethod
    def order() -> int:
        return 3


class ReciprocalMicrometers(ReciprocalLength):
    """um^-1."""

    @staticmethod
    def order() -> int:
        return 6
